/*
 ************************************************************************
 *
 *		   The lowest but one level file opener
 *
 * This code implements an "extended" system call open(2), which
 * opens a file for reading or writing. A function sys_open() defined
 * in this code takes the same arguments as open(2) and returns the
 * same result (that is, a file handle of an opened file, or -1 in
 * case of error). See "man 2 open" for more details.
 *
 *   int sys_open(const char * filename, const int mode, const int mask)
 *
 * If a filename argument given to sys_open() is a regular file/path
 * name, sys_open() is *identical* to open() (it simply exits
 * to open(), as a matter of fact). Unlike open() however, sys_open() can
 * handle extended file names like "cmd |" or "| cmd", where "cmd" is
 * anything that can be passed to /bin/sh. In that case, shell "/bin/sh"
 * is launched in a subprocess to interpret the "cmd"; the shell's stdin
 * or stdout become the file that is "opened" by sys_open().
 *
 * An extended file name "cmd |" assumes that cmd's standard output becomes
 * a "file" (a pipe, actually) the caller of sys_open() will read from.
 * That is, "cmd |" assumes a RDONLY open mode. By the same token,
 * a "| cmd" extended file name usually means that the caller of sys_open()
 * would then write into the file; all the data being written will be
 * passed to the stdin of the given "cmd". Still, _regardless_ of how
 * the extended file name is specified, "| cmd" or "cmd |", sys_open() always
 * obeys the opening mode as given by the 'mode' argument. But when
 * a file "cmd |" is open for writing, or "| cmd" is opened for reading,
 * sys_open() writes a gripe about it on the stderr. An extended file name
 * cannot be opened for reading _and_ writing; that is, O_RDWR open mode
 * is invalid with extended file names and causes result of -1 to be returned
 * (with errno set to EINVAL).
 * In the extended file names, leading spaces (before the '|' character)
 * and trainling spaces (after the '|' char) are allowed and ignored.
 *
 * This code is patterned after sys_open.cc in the earlier version
 * (v. 2.1) of c++advio distribution. The original inspiration for
 * _popen is due to
 *    Copyright (C) 1991, 1992 Per Bothner. (bothner@cygnus.com)
 *
 * $Id: sys_open.c,v 2.1 1998/03/29 23:22:18 oleg Exp oleg $
 *
 ************************************************************************
 */ 

#include <unistd.h>
#include <fcntl.h>
#include <sys/file.h>
#include <errno.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#define report_error(MSG,ARG) fprintf(stderr,MSG,(ARG))

/*
 *------------------------------------------------------------------------
 *
 * Launch a shell in a subprocess and have it interpret a string,
 * from cmd_beg up to (but not including) cmd_end. Shell's stdin or
 * stdout is directed to a pipe (depending on the 'mode' argument,
 * which can be either O_RDONLY or O_WRONLY). The other end of this
 * pipe is returned as the result of this function. In case of error,
 * the result is -1.
 * Note, fork() copies parent's address space. So it appears we may
 * modify cmd_beg and cmd_end (see below) at will without affecting
 * the parent. There is a hitch though: sys_open() might be called
 * with a constant string, like sys_open("cat < /tmp/a |"). In which
 * case the string is allocated in a BSS or even TEXT segment, which
 * is read-only. It remains read-only in the child process, so we
 * may not actually modify it.
 */

static int _popen(const char * cmd_beg, char * cmd_end, int mode)
{
  struct { int read_fd, write_fd; } pipe_fds;

  int parent_end, child_end;		/* ends of the pipe		*/
  int child_std_end;			/* File handles for stdin/out	*/
  int kid_id;
  
  assert( cmd_end > cmd_beg );
  if( pipe((int *)&pipe_fds) < 0 )
    return -1;

  if( mode == O_RDONLY )		/* We're reading, shell is writing */
    parent_end = pipe_fds.read_fd, child_end  = pipe_fds.write_fd,
    child_std_end = 1;			/* command's stdout handle	     */
  else					/* shell is reading, we're writing */
    parent_end = pipe_fds.write_fd, child_end  = pipe_fds.read_fd,
    child_std_end = 0;			/* command's stdin handle	     */

  if( (kid_id = fork()) == 0 )
  {					/* We're in kid's process	     */
					/* which is to execute the command   */
    char * cmd_string = malloc(cmd_end - cmd_beg + 1);
    strncpy(cmd_string,cmd_beg,cmd_end - cmd_beg);
    cmd_string[cmd_end - cmd_beg] = '\0';

    close(parent_end);			
    if( child_end != child_std_end )
    {
      dup2(child_end, child_std_end);
      close(child_end);
    }
    execl("/bin/sh", "sh", "-c", cmd_string, 0);
    _exit(127);				/* Executed only if execl failed!   */
  }

  close(child_end);			/* We're in the parent process	    */
  if( kid_id < 0 )
    close(parent_end), parent_end = -1;	/* if fork failed           */

  return parent_end;
}

/*
 *------------------------------------------------------------------------
 *			  An extended 'open(2)'
 */
	/* if str begins with a pipe char '|' (after possibly several spaces)
	 * return a pointer to the character right after that.
	 * Otherwise, return NULL
	 */
static const char * check_leading_barchar(const char * str)
{
  register const char * p = str;
  while( *p == ' ' )
    p++;
  return *p == '|' ? p+1 : (char *)0;
}

	/* if str ends with a pipe char '|' (followed by possibly several
	 * spaces) return a pointer to it.
	 * Otherwise, return NULL
	 */
static const char * check_trailing_barchar(const char * str)
{
  register const char * p = str + strlen(str);
  while( *--p == ' ' && p > str )
    ;
  return *p == '|' ? p : (char *)0;
}

int sys_open(const char *filename, const int mode, const int mask)
{
  register const char *p = check_leading_barchar(filename);
  if( p != (char*)0 )
  {				/* fname starts with '|'		*/
  				/* p points to the first char after |   */
    if( *p == '\0' )
      return (errno = EINVAL), -1;	/* Empty command		*/
      
    switch( mode & (O_WRONLY | O_RDONLY) )
    {
      case O_RDONLY:
      	   report_error
	   ("File name '%s' looks like the pipe to write to,"
	    "\nbut the open mode is not WRITE_ONLY\n",filename);
   	   return _popen(p,filename+strlen(filename),O_RDONLY);
   	   
      case O_WRONLY:
   	   return _popen(p,filename+strlen(filename),O_WRONLY);
   	   
      default:
      	   report_error
	   ("File name '%s' looks like the pipe to write to,"
	    "\nbut the open mode is not WRITE_ONLY nor READ_ONLY\n",filename);
	   errno = EINVAL;
	   return -1;
    }
  }
				/* '|' is the last char of the filename */
  else if( (p=check_trailing_barchar(filename)) != (char *)0 )
  {
    switch( mode & (O_WRONLY | O_RDONLY) )
    {
      case O_RDONLY:
   	   return _popen(filename,p,O_RDONLY);
   	   
      case O_WRONLY:
      	   report_error
	   ("File name '%s' looks like the pipe to read from,"
	    "\nbut the open mode is not READ_ONLY\n",filename);
   	   return _popen(filename,p,O_WRONLY);
   	   
      default:
      	   report_error
	   ("File name '%s' looks like the pipe to read from,"
	    "\nbut the open mode is not WRITE_ONLY nor READ_ONLY\n",filename);
	   errno = EINVAL;
	   return -1;
    }
  }
  else
    return open(filename,mode,mask);

  return -1;		/* Unnecessary, but gcc really likes it... */
}

